@@ -43,10 +43,10 @@
 
 struct _xlator;
 typedef struct _xlator xlator_t;
-struct _dir_entry_t;
-typedef struct _dir_entry_t dir_entry_t;
-struct _gf_dirent_t;
-typedef struct _gf_dirent_t gf_dirent_t;
+struct _dir_entry;
+typedef struct _dir_entry dir_entry_t;
+struct _gf_dirent;
+typedef struct _gf_dirent gf_dirent_t;
 struct _loc;
 typedef struct _loc loc_t;
 
@@ -58,8 +58,6 @@ typedef int32_t (*event_notify_fn_t) (xlator_t *this, int32_t event, void *data,
 #include "gf-dirent.h"
 #include "stack.h"
 #include "iobuf.h"
-#include "inode.h"
-#include "fd.h"
 #include "globals.h"
 #include "iatt.h"
 #include "options.h"
@@ -467,6 +465,22 @@ typedef int32_t (*fop_setactivelk_cbk_t) (call_frame_t *frame, void *cookie,
                                             xlator_t *this, int32_t op_ret,
                                             int32_t op_errno, dict_t *xdata);
 
+typedef int32_t (*fop_put_cbk_t) (call_frame_t *frame, void *cookie,
+                                  xlator_t *this, int32_t op_ret,
+                                  int32_t op_errno, inode_t *inode,
+                                  struct iatt *buf, struct iatt *preparent,
+                                  struct iatt *postparent, dict_t *xdata);
+
+typedef int32_t (*fop_icreate_cbk_t) (call_frame_t *frame, void *cookie,
+                                      xlator_t *this, int32_t op_ret,
+                                      int32_t op_errno, inode_t *inode,
+                                      struct iatt *buf, dict_t *xdata);
+
+typedef int32_t (*fop_namelink_cbk_t) (call_frame_t *frame, void *cookie,
+                                       xlator_t *this, int32_t op_ret,
+                                       int32_t op_errno, struct iatt *prebuf,
+                                       struct iatt *postbuf, dict_t *xdata);
+
 typedef int32_t (*fop_lookup_t) (call_frame_t *frame,
                                  xlator_t *this,
                                  loc_t *loc,
@@ -727,13 +741,23 @@ typedef int32_t (*fop_setactivelk_t) (call_frame_t *frame, xlator_t *this,
                                       lock_migration_info_t *locklist,
                                       dict_t *xdata);
 
+typedef int32_t (*fop_put_t) (call_frame_t *frame, xlator_t *this, loc_t *loc,
+                              mode_t mode, mode_t umask, uint32_t flags,
+                              struct iovec *vector, int32_t count, off_t offset,
+                              struct iobref *iobref, dict_t *xattr,
+                              dict_t *xdata);
+
+typedef int32_t (*fop_icreate_t) (call_frame_t *frame, xlator_t *this,
+                                  loc_t *loc, mode_t mode, dict_t *xdata);
+
+typedef int32_t (*fop_namelink_t) (call_frame_t *frame, xlator_t *this,
+                                   loc_t *loc, dict_t *xdata);
+
+/* WARNING: make sure the list is in order with FOP definition in
+   `rpc/xdr/src/glusterfs-fops.x`.
+   If it is not in order, mainly the metrics related feature would be broken */
 struct xlator_fops {
-        fop_lookup_t         lookup;
         fop_stat_t           stat;
-        fop_fstat_t          fstat;
-        fop_truncate_t       truncate;
-        fop_ftruncate_t      ftruncate;
-        fop_access_t         access;
         fop_readlink_t       readlink;
         fop_mknod_t          mknod;
         fop_mkdir_t          mkdir;
@@ -742,51 +766,62 @@ struct xlator_fops {
         fop_symlink_t        symlink;
         fop_rename_t         rename;
         fop_link_t           link;
-        fop_create_t         create;
+        fop_truncate_t       truncate;
         fop_open_t           open;
         fop_readv_t          readv;
         fop_writev_t         writev;
+        fop_statfs_t         statfs;
         fop_flush_t          flush;
         fop_fsync_t          fsync;
-        fop_opendir_t        opendir;
-        fop_readdir_t        readdir;
-        fop_readdirp_t       readdirp;
-        fop_fsyncdir_t       fsyncdir;
-        fop_statfs_t         statfs;
         fop_setxattr_t       setxattr;
         fop_getxattr_t       getxattr;
-        fop_fsetxattr_t      fsetxattr;
-        fop_fgetxattr_t      fgetxattr;
         fop_removexattr_t    removexattr;
-        fop_fremovexattr_t   fremovexattr;
+        fop_opendir_t        opendir;
+        fop_fsyncdir_t       fsyncdir;
+        fop_access_t         access;
+        fop_create_t         create;
+        fop_ftruncate_t      ftruncate;
+        fop_fstat_t          fstat;
         fop_lk_t             lk;
+        fop_lookup_t         lookup;
+        fop_readdir_t        readdir;
         fop_inodelk_t        inodelk;
         fop_finodelk_t       finodelk;
         fop_entrylk_t        entrylk;
         fop_fentrylk_t       fentrylk;
-        fop_rchecksum_t      rchecksum;
         fop_xattrop_t        xattrop;
         fop_fxattrop_t       fxattrop;
+        fop_fgetxattr_t      fgetxattr;
+        fop_fsetxattr_t      fsetxattr;
+        fop_rchecksum_t      rchecksum;
         fop_setattr_t        setattr;
         fop_fsetattr_t       fsetattr;
+        fop_readdirp_t       readdirp;
+
+        /* These 3 are required to keep the index same as GF_FOP_##FOP */
+        void *forget_placeholder;
+        void *release_placeholder;
+        void *releasedir_placeholder;
+
         fop_getspec_t        getspec;
-	fop_fallocate_t	     fallocate;
-	fop_discard_t	     discard;
+        fop_fremovexattr_t   fremovexattr;
+        fop_fallocate_t	     fallocate;
+        fop_discard_t	     discard;
         fop_zerofill_t       zerofill;
         fop_ipc_t            ipc;
         fop_seek_t           seek;
         fop_lease_t          lease;
         fop_compound_t       compound;
-        fop_getactivelk_t   getactivelk;
-        fop_setactivelk_t  setactivelk;
+        fop_getactivelk_t    getactivelk;
+        fop_setactivelk_t    setactivelk;
+        fop_put_t            put;
+        fop_icreate_t        icreate;
+        fop_namelink_t       namelink;
 
         /* these entries are used for a typechecking hack in STACK_WIND _only_ */
-        fop_lookup_cbk_t         lookup_cbk;
+        /* make sure to add _cbk variables only after defining regular fops as
+           its relative position is used to get the index */
         fop_stat_cbk_t           stat_cbk;
-        fop_fstat_cbk_t          fstat_cbk;
-        fop_truncate_cbk_t       truncate_cbk;
-        fop_ftruncate_cbk_t      ftruncate_cbk;
-        fop_access_cbk_t         access_cbk;
         fop_readlink_cbk_t       readlink_cbk;
         fop_mknod_cbk_t          mknod_cbk;
         fop_mkdir_cbk_t          mkdir_cbk;
@@ -795,43 +830,57 @@ struct xlator_fops {
         fop_symlink_cbk_t        symlink_cbk;
         fop_rename_cbk_t         rename_cbk;
         fop_link_cbk_t           link_cbk;
-        fop_create_cbk_t         create_cbk;
+        fop_truncate_cbk_t       truncate_cbk;
         fop_open_cbk_t           open_cbk;
         fop_readv_cbk_t          readv_cbk;
         fop_writev_cbk_t         writev_cbk;
+        fop_statfs_cbk_t         statfs_cbk;
         fop_flush_cbk_t          flush_cbk;
         fop_fsync_cbk_t          fsync_cbk;
-        fop_opendir_cbk_t        opendir_cbk;
-        fop_readdir_cbk_t        readdir_cbk;
-        fop_readdirp_cbk_t       readdirp_cbk;
-        fop_fsyncdir_cbk_t       fsyncdir_cbk;
-        fop_statfs_cbk_t         statfs_cbk;
         fop_setxattr_cbk_t       setxattr_cbk;
         fop_getxattr_cbk_t       getxattr_cbk;
-        fop_fsetxattr_cbk_t      fsetxattr_cbk;
-        fop_fgetxattr_cbk_t      fgetxattr_cbk;
         fop_removexattr_cbk_t    removexattr_cbk;
-        fop_fremovexattr_cbk_t   fremovexattr_cbk;
+        fop_opendir_cbk_t        opendir_cbk;
+        fop_fsyncdir_cbk_t       fsyncdir_cbk;
+        fop_access_cbk_t         access_cbk;
+        fop_create_cbk_t         create_cbk;
+        fop_ftruncate_cbk_t      ftruncate_cbk;
+        fop_fstat_cbk_t          fstat_cbk;
         fop_lk_cbk_t             lk_cbk;
+        fop_lookup_cbk_t         lookup_cbk;
+        fop_readdir_cbk_t        readdir_cbk;
         fop_inodelk_cbk_t        inodelk_cbk;
         fop_finodelk_cbk_t       finodelk_cbk;
         fop_entrylk_cbk_t        entrylk_cbk;
         fop_fentrylk_cbk_t       fentrylk_cbk;
-        fop_rchecksum_cbk_t      rchecksum_cbk;
         fop_xattrop_cbk_t        xattrop_cbk;
         fop_fxattrop_cbk_t       fxattrop_cbk;
+        fop_fgetxattr_cbk_t      fgetxattr_cbk;
+        fop_fsetxattr_cbk_t      fsetxattr_cbk;
+        fop_rchecksum_cbk_t      rchecksum_cbk;
         fop_setattr_cbk_t        setattr_cbk;
         fop_fsetattr_cbk_t       fsetattr_cbk;
+        fop_readdirp_cbk_t       readdirp_cbk;
+
+        /* These 3 are required to keep the index same as GF_FOP_##FOP */
+        void *forget_placeholder_cbk;
+        void *release_placeholder_cbk;
+        void *releasedir_placeholder_cbk;
+
         fop_getspec_cbk_t        getspec_cbk;
-	fop_fallocate_cbk_t	 fallocate_cbk;
-	fop_discard_cbk_t	 discard_cbk;
+        fop_fremovexattr_cbk_t   fremovexattr_cbk;
+        fop_fallocate_cbk_t      fallocate_cbk;
+        fop_discard_cbk_t        discard_cbk;
         fop_zerofill_cbk_t       zerofill_cbk;
         fop_ipc_cbk_t            ipc_cbk;
         fop_seek_cbk_t           seek_cbk;
         fop_lease_cbk_t          lease_cbk;
         fop_compound_cbk_t       compound_cbk;
-        fop_getactivelk_cbk_t   getactivelk_cbk;
-        fop_setactivelk_cbk_t  setactivelk_cbk;
+        fop_getactivelk_cbk_t    getactivelk_cbk;
+        fop_setactivelk_cbk_t    setactivelk_cbk;
+        fop_put_cbk_t            put_cbk;
+        fop_icreate_cbk_t        icreate_cbk;
+        fop_namelink_cbk_t       namelink_cbk;
 };
 
 typedef int32_t (*cbk_forget_t) (xlator_t *this,
@@ -906,6 +955,10 @@ typedef struct xlator_list {
         struct xlator_list *next;
 } xlator_list_t;
 
+typedef struct fop_metrics {
+        gf_atomic_t fop;
+        gf_atomic_t cbk; /* only updaed when there is failure */
+} fop_metrics_t;
 
 struct _xlator {
         /* Built during parsing */
@@ -929,12 +982,30 @@ struct _xlator {
         int32_t           (*init) (xlator_t *this);
         int32_t           (*reconfigure) (xlator_t *this, dict_t *options);
 	int32_t           (*mem_acct_init) (xlator_t *this);
+	int32_t           (*dump_metrics) (xlator_t *this, int fd);
+
 	event_notify_fn_t notify;
 
         gf_loglevel_t    loglevel;   /* Log level for translator */
 
-        /* for latency measurement */
-        fop_latency_t latencies[GF_FOP_MAXVALUE];
+        struct {
+                int64_t  client_latency; /* This is in 'milliseconds' units */
+                struct {
+                        /* for latency measurement */
+                        fop_metrics_t metrics[GF_FOP_MAXVALUE];
+
+                        gf_atomic_t count;
+                } total;
+
+                struct {
+                        /* for latency measurement */
+                        fop_latency_t latencies[GF_FOP_MAXVALUE];
+                        /* for latency measurement */
+                        fop_metrics_t metrics[GF_FOP_MAXVALUE];
+
+                        gf_atomic_t count;
+                } interval;
+        } stats;
 
         /* Misc */
         eh_t               *history; /* event history context */
@@ -956,6 +1027,18 @@ struct _xlator {
 
         /* Its used as an index to inode_ctx*/
         uint32_t            xl_id;
+
+        /* op_version: initialized in xlator code itself */
+        uint32_t op_version[GF_MAX_RELEASES];
+
+        /* flags: initialized in xlator code itself */
+        uint32_t flags;
+
+        /* id: unique, initialized in xlator code itself */
+        uint32_t id;
+
+        /* identifier: a full string which can unique identify the xlator */
+        char *identifier;
 };
 
 typedef struct {
@@ -966,6 +1049,71 @@ typedef struct {
         event_notify_fn_t       notify;
 } class_methods_t;
 
+/* This would be the only structure which needs to be exported by
+   the translators. For the backward compatibility, in 4.x series
+   even the old exported fields will be supported */
+typedef struct {
+        /* op_version: will be used by volume generation logic to figure
+           out whether to insert it in graph or no, based on cluster's
+           operating version.
+           default value: 0, which means good to insert always */
+        uint32_t op_version[GF_MAX_RELEASES];
+
+        /* flags: will be used by volume generation logic to optimize the
+           placements etc.
+           default value: 0, which means don't treat it specially */
+        uint32_t flags;
+
+        /* xlator_id: unique per xlator. make sure to have no collission
+           in this ID */
+        uint32_t xlator_id;
+
+        /* identifier: a string constant */
+        char *identifier;
+
+        /* struct options: if the translator takes any 'options' from the
+           volume file, then that should be defined here. optional. */
+        volume_option_t *options;
+
+        /* init(): mandatory method, will be called during the
+           graph initialization */
+        int32_t (*init) (xlator_t *this);
+
+        /* fini(): optional method, will be initialized to default
+           method which would just free the 'xlator->private' variable.
+           This method is called when the graph is no more in use, and
+           is being destroyed. Also when SIGTERM is received */
+        void (*fini) (xlator_t *this);
+
+        /* reconfigure(): optional method, will be initialized to default
+           method in case not provided by xlator. This method is called
+           when there are only option changes in xlator, and no graph change.
+           eg., a 'gluster volume set' command */
+        int32_t (*reconfigure) (xlator_t *this, dict_t *options);
+
+        /* mem_acct_init(): used for memory accounting inside of the xlator.
+           optional. called during translator initialization */
+	int32_t (*mem_acct_init) (xlator_t *this);
+
+        /* dump_metrics(): used for providing internal metrics. optional */
+	int32_t (*dump_metrics) (xlator_t *this, int fd);
+
+        /* notify(): used for handling the notification of events from either
+           the parent or child in the graph. optional. */
+        event_notify_fn_t notify;
+
+        /* struct fops: mandatory. provides all the filesystem operations
+           methods of the xlator */
+        struct xlator_fops *fops;
+        /* struct cbks: optional. provides methods to handle
+           inode forgets, and fd releases */
+        struct xlator_cbks *cbks;
+
+        /* dumpops: a structure again, with methods to dump the details.
+           optional. */
+        struct xlator_dumpops  *dumpops;
+} xlator_api_t;
+
 #define xlator_has_parent(xl) (xl->parents != NULL)
 
 #define XLATOR_NOTIFY(_xl, params ...)          \
@@ -1029,6 +1177,7 @@ void loc_pargfid (loc_t *loc, uuid_t pargfid);
 char* loc_gfid_utoa (loc_t *loc);
 gf_boolean_t loc_is_root (loc_t *loc);
 int32_t loc_build_child (loc_t *child, loc_t *parent, char *name);
+gf_boolean_t loc_is_nameless (loc_t *loc);
 int xlator_mem_acct_init (xlator_t *xl, int num_types);
 int is_gf_log_command (xlator_t *trans, const char *name, char *value);
 int glusterd_check_log_level (const char *value);
@@ -1041,8 +1190,11 @@ enum gf_hdsk_event_notify_op {
 gf_boolean_t
 is_graph_topology_equal (glusterfs_graph_t *graph1, glusterfs_graph_t *graph2);
 int
-glusterfs_volfile_reconfigure (int oldvollen, FILE *newvolfile_fp,
-                               glusterfs_ctx_t *ctx, const char *oldvolfile);
+glusterfs_volfile_reconfigure (FILE *newvolfile_fp, glusterfs_ctx_t *ctx);
+
+int
+gf_volfile_reconfigure (int oldvollen, FILE *newvolfile_fp,
+                        glusterfs_ctx_t *ctx, const char *oldvolfile);
 
 int
 loc_touchup (loc_t *loc, const char *name);
@@ -1061,4 +1213,8 @@ void xlator_init_unlock (void);
 int
 copy_opts_to_child (xlator_t *src, xlator_t *dst, char *glob);
 
+int
+glusterfs_delete_volfile_checksum (glusterfs_ctx_t *ctx,
+                                   const char *volfile_id);
+
 #endif /* _XLATOR_H */
